#!/usr/bin/env python3
#
# Sniffles2
# A fast structural variant caller for long-read sequencing data
#
# Created:     27.08.2021
# Author:      Moritz Smolka
# Maintainer:  Hermann Romanek
# Contact:     sniffles@romanek.at
#
import copy
import gc
import logging
import math
import multiprocessing
import os
import threading
import time
from argparse import Namespace
from collections import deque
from dataclasses import dataclass
from typing import Optional, Union, Callable

import pysam

from sniffles import cluster
from sniffles import leadprov
from sniffles import postprocessing
from sniffles import snf
from sniffles import sv
from sniffles.region import Region
from sniffles.result import Result, ErrorResult, CallResult, GenotypeResult, CombineResult
from sniffles.snfp import PopulationSNF


@dataclass
class Task:
    """
    A task is a generic unit of work sent to a child process to be worked on in parallel. Must be pickleable.
    """
    id: int
    sv_id: int
    contig: str
    start: int
    end: int
    config: Namespace
    assigned_process_id: Optional[int] = None
    lead_provider: leadprov.LeadProvider = None
    bam: object = None
    tandem_repeats: list = None
    genotype_svs: list = None
    regions: list[Region] = None
    _logger = None
    result: Result = None

    def __str__(self):
        return f'Task #{self.id}'

    @property
    def logger(self) -> logging.Logger:
        if self._logger is None:
            self._logger = logging.getLogger(f'sniffles.progress')

        return self._logger

    @property
    def done(self) -> bool:
        return self.result is not None

    @property
    def success(self) -> bool:
        return self.done and not self.result.error

    def add_result(self, result: Result) -> None:
        self.result = result

    def execute(self, worker: 'SnifflesWorker' = None) -> Optional[Result]:
        """
        Execute this Task, returning a Result object
        :param worker is the worker executing this task
        """
        raise NotImplemented

    def build_leadtab(self):
        assert (self.lead_provider is None)

        config = self.config

        if config.input_is_cram and config.reference is not None:
            self.bam = pysam.AlignmentFile(config.input, config.input_mode, require_index=True, reference_filename=config.reference)
        else:
            self.bam = pysam.AlignmentFile(config.input, config.input_mode, require_index=True)
        self.lead_provider = leadprov.LeadProvider(config, self.id * config.task_read_id_offset_mult)
        externals = self.lead_provider.build_leadtab(self.regions if self.regions else [Region(self.contig, self.start, self.end)], self.bam)
        return externals, self.lead_provider.read_count

    def call_candidates(self, keep_qc_fails, config) -> list[sv.SVCall]:
        candidates = []
        for svtype in sv.TYPES:
            for svcluster in cluster.resolve(svtype, self.lead_provider, config, self.tandem_repeats):
                for svcall in sv.call_from(svcluster, config, keep_qc_fails, self):
                    if config.dev_trace_read is not False:
                        cluster_has_read = False
                        for ld in svcluster.leads:
                            if ld.read_qname == config.dev_trace_read:
                                cluster_has_read = True
                        if cluster_has_read:
                            import copy
                            svcall_copy = copy.deepcopy(svcall)
                            svcall_copy.postprocess = None
                            print(f"[DEV_TRACE_READ] [3/4] [Task.call_candidates] Read {config.dev_trace_read} -> Cluster {svcluster.id} -> preliminary SVCall {svcall_copy}")
                    candidates.append(svcall)

        self.coverage_average_fwd, self.coverage_average_rev = postprocessing.coverage(candidates, self.lead_provider, config)
        self.coverage_average_total = self.coverage_average_fwd + self.coverage_average_rev
        return candidates

    def finalize_candidates(self, candidates: list['SVCall'], keep_qc_fails, config):
        passed = []
        for svcall in candidates:
            svcall.qc = svcall.qc and postprocessing.qc_sv(svcall, config)
            if not keep_qc_fails and not svcall.qc:
                continue
            svcall.qc = svcall.qc and postprocessing.qc_sv_support(svcall, self.coverage_average_total, config)
            if not keep_qc_fails and not svcall.qc:
                continue

            postprocessing.annotate_sv(svcall, config)

            svcall.qc = svcall.qc and postprocessing.qc_sv_post_annotate(svcall, config)

            if config.dev_trace_read:
                cluster_has_read = False
                for ld in svcall.postprocess.cluster.leads:
                    if ld.read_qname == config.dev_trace_read:
                        cluster_has_read = True
                if cluster_has_read:
                    import copy
                    svcall_copy = copy.deepcopy(svcall)
                    svcall_copy.postprocess = None
                    print(f"[DEV_TRACE_READ] [4/4] [Task.finalize_candidates] Read {config.dev_trace_read} -> Cluster {svcall.postprocess.cluster.id} -> finalized SVCall, QC={svcall_copy.qc}: {svcall_copy}")

            if not keep_qc_fails and not svcall.qc:
                continue

            svcall.finalize()  # Remove internal information (not written to output) before sending to mainthread for VCF writing
            passed.append(svcall)
        return passed


class CallTask(Task):
    """
    """

    def execute(self, worker: 'SnifflesWorker' = None) -> CallResult:
        config = self.config

        if config.snf is not None or config.no_qc:
            qc = False
        else:
            qc = True

        _, read_count = self.build_leadtab()
        svcandidates = self.call_candidates(qc, config)
        svcalls = self.finalize_candidates(svcandidates, not qc, config)
        if not config.no_qc:
            svcalls = [s for s in svcalls if s.qc]

        if config.sort:
            svcalls = sorted(svcalls, key=lambda svcall: svcall.pos)

        from sniffles.result import CallResult
        result = CallResult(self, svcalls, read_count)

        if config.snf is not None:  # and len(svcandidates):
            snf_filename = f"{config.snf}.tmp_{self.id}.snf"

            with open(snf_filename, "wb") as handle:
                snf_out = snf.SNFile(config, handle)
                for cand in svcandidates:
                    snf_out.store(cand)
                snf_out.annotate_block_coverages(self.lead_provider)
                snf_out.write_and_index()
                handle.close()
            result.snf_filename = snf_filename
            result.snf_index = snf_out.get_index()
            result.snf_total_length = snf_out.get_total_length()
            result.snf_candidate_count = len(svcandidates)
            result.has_snf = True

        result.coverage_average_total = self.coverage_average_total

        return result


class GenotypeTask(Task):
    def execute(self, worker: 'SnifflesWorker' = None) -> Optional[GenotypeResult]:
        config = self.config

        qc = False
        _, read_count = self.build_leadtab()
        svcandidates = self.call_candidates(qc, config=config)
        svcalls = self.finalize_candidates(svcandidates, not qc, config=config)

        binsize = 5000
        binedge = int(binsize / 10)
        genotype_svs_svtypes_bins = {svtype: {} for svtype in sv.TYPES}
        for genotype_sv in self.genotype_svs:
            genotype_sv.genotype_match_sv = None
            genotype_sv.genotype_match_dist = math.inf

            if genotype_sv.svtype not in genotype_svs_svtypes_bins:
                logging.getLogger('sniffles').warning(f'Unsupported SVTYPE: {genotype_sv.svtype}')
                continue

            bins = [int(genotype_sv.pos / binsize) * binsize]
            if genotype_sv.pos % binsize < binedge:
                bins.append((int(genotype_sv.pos / binsize) - 1) * binsize)
            if genotype_sv.pos % binsize > binsize - binedge:
                bins.append((int(genotype_sv.pos / binsize) + 1) * binsize)

            for bin in bins:
                if bin not in genotype_svs_svtypes_bins[genotype_sv.svtype]:
                    genotype_svs_svtypes_bins[genotype_sv.svtype][bin] = []
                genotype_svs_svtypes_bins[genotype_sv.svtype][bin].append(genotype_sv)

        for cand in svcandidates:
            bin = int(cand.pos / binsize) * binsize
            if bin not in genotype_svs_svtypes_bins[cand.svtype]:
                continue
            if cand.svtype == "BND":
                for genotype_sv in genotype_svs_svtypes_bins[cand.svtype][bin]:
                    dist = abs(genotype_sv.pos - cand.pos)
                    # if minlen>0 and dist < genotype_sv.genotype_match_dist and dist <= config.cluster_merge_bnd * 2:
                    if dist < genotype_sv.genotype_match_dist and dist <= config.cluster_merge_bnd:
                        if cand.bnd_info.mate_contig == genotype_sv.bnd_info.mate_contig:
                            genotype_sv.genotype_match_sv = cand
                            genotype_sv.genotype_match_dist = dist
            else:
                for genotype_sv in genotype_svs_svtypes_bins[cand.svtype][bin]:
                    dist = abs(genotype_sv.pos - cand.pos) + abs(abs(genotype_sv.svlen) - abs(cand.svlen))
                    minlen = float(min(abs(genotype_sv.svlen), abs(cand.svlen)))
                    if minlen > 0 and dist < genotype_sv.genotype_match_dist and dist <= config.combine_match * math.sqrt(
                            minlen) and dist <= config.combine_match_max:
                        genotype_sv.genotype_match_sv = cand
                        genotype_sv.genotype_match_dist = dist

        postprocessing.coverage(self.genotype_svs, self.lead_provider, config)

        # Determine genotypes for unmatched input SVs
        for svcall in self.genotype_svs:
            coverage_list = [svcall.coverage_start, svcall.coverage_center, svcall.coverage_end]
            coverage_list = [c for c in coverage_list if c is not None]
            if len(coverage_list) == 0:
                return
            coverage = round(sum(coverage_list) / len(coverage_list))
            svcall.genotypes = {}
            if coverage > 0:
                svcall.genotypes[0] = (0, 0, 0, coverage, 0, (None, None))
            else:
                svcall.genotypes[0] = config.genotype_none

        from sniffles.result import GenotypeResult
        return GenotypeResult(self, self.genotype_svs, read_count)


class CombineTask(Task):
    """
    Task to merge/combine multiple SNF files into one.
    """
    # target number of blocks to process in one task. this is the total number of blocks over all input files,
    # i.e. merging 100 files means 100 blocks wrt this value
    TARGET_WORK_PER_TASK = 10000

    result_class = CombineResult

    block_indices: list[int] = None  # List of block starts to process on this task

    def __init__(self, *args, **kwargs):
        self.result_class = kwargs.pop('result_class', None) or self.result_class
        super().__init__(*args, **kwargs)
        self.generate_blocks()

    def generate_blocks(self):
        """
        Generate a set of blocks
        """
        if self.regions:
            block_indices = set()
            for r in self.regions:
                start = r.start // self.config.snf_block_size * self.config.snf_block_size
                block_indices |= set(range(start, r.end + self.config.snf_block_size, self.config.snf_block_size))
            self.block_indices = list(sorted(block_indices))
        else:
            self.block_indices = list(range(self.start, self.end + self.config.snf_block_size, self.config.snf_block_size))

    def __str__(self):
        if len(self.block_indices) > 0:
            return f'''Task {self.id} Contig {self.contig} [{self.start} ({self.block_indices[0]}) .. {self.end} ({self.block_indices[-1]})]'''
        else:
            return f'Task {self.id} [no blocks available]'

    def clone(self, first_block: int, block_count: int, new_id: int = None, emit_first: bool = True) -> 'CombineTask':
        """
        Clone this task with a subset of the blocks to be processed.
        """
        obj = copy.copy(self)
        if new_id is not None:
            obj.id = new_id
        obj.block_indices = self.block_indices[first_block:first_block + block_count]
        obj.start = obj.block_indices[0]
        obj.end = obj.block_indices[-1] + obj.config.snf_block_size
        return obj

    def scatter(self) -> list['CombineTask']:
        """
        Scatter this task to a number of tasks to be run in parallel, on block level:
        - Distribute blocks equally to tasks, each one working on consecutive blocks
        - First task gets any blocks that can not be evenly distributed
        - Tasks other than the first one have to process their first block for kept groups, and
          should not emit calls for this block (as these will be emitted by the previous task)
        """
        total_blocks = len(self.block_indices) * len(self.config.sample_ids_vcf)
        if total_blocks <= self.TARGET_WORK_PER_TASK or self.config.threads <= 1:
            return [self]

        blocks_per_task = (total_blocks // self.TARGET_WORK_PER_TASK)

        return [
            self.clone(
                fb,
                blocks_per_task,
                new_id=self.id + i + 1
            ) for i, fb in enumerate(range(0, len(self.block_indices), blocks_per_task))
        ]

    def execute(self, worker: 'SnifflesWorker' = None):
        samples_headers_snf = {}
        for snf_info in self.config.snf_input_info:
            snf_in = snf.SNFile(self.config, open(snf_info["filename"], "rb"), filename=snf_info["filename"])
            snf_in.read_header()
            samples_headers_snf[snf_info["internal_id"]] = snf_in

            if self.config.combine_close_handles:
                snf_in.close()

        if self.config.combine_population:
            self.config.combine_population = PopulationSNF.open(self.config.combine_population)

        result = self.result_class(self, [], 0)

        # block_groups_keep_threshold=5000
        # TODO: Parameterize
        bin_min_size = self.config.combine_min_size
        bin_max_candidates = max(25, int(len(self.config.snf_input_info) * 0.5))
        overlap_abs = self.config.combine_overlap_abs
        support_threshold = self.config.combine_support_threshold

        sample_internal_ids = set(samples_headers_snf.keys())

        #
        # Load candidate SVs from all samples for each block separately and cluster them based on start position
        #
        candidates_processed = 0
        groups_keep = {svtype: list() for svtype in sv.TYPES}
        calls = []

        for cur, block_index in enumerate(self.block_indices):  # iterate over all blocks
            self.logger.info(f'Processing block {cur + 1}/{len(self.block_indices)} (active calls: {sv.SVCall._counter} groups: {sv.SVGroup._counter})')
            samples_blocks = {}
            if calls:
                result.store_calls(calls)
                calls = []

            for sample_internal_id, sample_snf in samples_headers_snf.items():
                blocks = sample_snf.read_blocks(self.contig, block_index)
                samples_blocks[sample_internal_id] = blocks
                # sample_internal_id is the number of the processed file
                # blocks is a list[dict[str, list[SVCall]]] {'INS': [...], 'DEL': [...], ...}

            for svtype in sv.TYPES:
                bins = {}
                for sample_internal_id, sample_snf in samples_headers_snf.items():  # fetch current block for each file
                    blocks = samples_blocks[sample_internal_id]
                    reqc = sample_snf.reqc

                    if blocks is None:
                        continue
                    for block in blocks:  # usually only 1 block
                        for cand in block[svtype]:
                            # if config.combine_pass_only and (cand.qc==False or cand.filter!="PASS"):
                            #    continue
                            if cand.support < support_threshold:
                                continue

                            if reqc:
                                postprocessing.genotype_sv(cand, self.config)

                            cand.sample_internal_id = sample_internal_id

                            bin = int(cand.pos / bin_min_size) * bin_min_size
                            if bin not in bins:
                                bins[bin] = [cand]
                            else:
                                bins[bin].append(cand)
                        candidates_processed += len(block[svtype])

                if len(bins) == 0:
                    continue

                size = 0
                svcands = []
                keep = groups_keep[svtype]
                sorted_bins = sorted(bins)
                last_bin = sorted_bins[-1]
                for curr_bin in sorted_bins:
                    svcands.extend(bins[curr_bin])  # here SVCalls from bins are collected...
                    size += bin_min_size

                    if (not self.config.combine_exhaustive and len(svcands) >= bin_max_candidates) or curr_bin == last_bin:
                        if len(svcands) == 0:
                            size = 0
                            continue

                        svgroups = cluster.resolve_block_groups(svtype, svcands, keep, self.config)
                        groups_call = []
                        keep = []
                        for group in svgroups:
                            coverage_bin = int(
                                group.pos_mean / self.config.coverage_binsize_combine) * self.config.coverage_binsize_combine
                            # High Intensity loop
                            for non_included_sample in sample_internal_ids - group.included_samples:
                                if samples_blocks[non_included_sample] is not None and coverage_bin in samples_blocks[non_included_sample][0]["_COVERAGE"]:
                                    coverage = samples_blocks[non_included_sample][0]["_COVERAGE"][coverage_bin]
                                else:
                                    coverage = 0
                                if non_included_sample in group.coverages_nonincluded:
                                    group.coverages_nonincluded[non_included_sample] = max(
                                        coverage,
                                        group.coverages_nonincluded[non_included_sample]
                                    )
                                else:
                                    group.coverages_nonincluded[non_included_sample] = coverage

                            if abs(group.pos_mean - curr_bin) < max(size * 0.5, overlap_abs):
                                keep.append(group)
                            else:
                                groups_call.append(group)

                        calls.extend(sv.call_groups(groups_call, self.config, self))

                        size = 0
                        svcands = []

                groups_keep[svtype] = keep

        for svtype in groups_keep:
            calls.extend(sv.call_groups(groups_keep[svtype], self.config, self))

        if calls:
            result.store_calls(calls)

        result.finalize()
        return result


class ShutdownTask:
    id = None

    def __str__(self):
        return 'Shutdown Request'

    def execute(self, *args, **kwargs) -> Result:
        raise SnifflesWorker.Shutdown


class SnifflesWorker:
    """
    Handle for a worker process. Since we're forking, this class will be available in
    both the parent and the worker processes.
    """
    id: int  # sequential ID of this worker, starting with 0 for the first
    externals: list = None
    recycle: bool = False
    running = True
    pid: int = None
    # Event to shut down heartbeat threads
    _shutdown: threading.Event
    _heartbeat: float = 0  # last heartbeat received
    HEARTBEAT_INTERVAL = 3  # in seconds
    HEARTBEAT_TIMEOUT = 10  # in seconds

    class Shutdown(Exception):
        """
        Indicates this worker process should shut down
        """

    def __init__(self, process_id: int, config: Namespace, tasks: deque[Task], recycle_hint: Union[bool, Callable] = None):
        self.id = process_id
        self.config = config
        self.tasks = tasks
        self.task = None
        self.finished_tasks = []
        self.recycle = recycle_hint

        self.pipe_main, self.pipe_worker = multiprocessing.Pipe()
        self.heartbeat_main, self.heartbeat_worker = multiprocessing.Pipe()

        self.process = multiprocessing.Process(
            target=self.run_worker,
            daemon=True
        )

        self._logger = logging.getLogger('sniffles.worker')

    def __str__(self):
        return f'Worker {self.id} @ process {self.pid}'

    def start(self) -> None:
        self._logger.info(f'Starting worker {self.id}')
        self.running = True
        self.process.start()
        self._heartbeat = time.monotonic()

    def maybe_recycle(self):
        """
        Recycle this worker if that has been requested
        """
        recycle = self.recycle(self.id, self.process.pid) if callable(self.recycle) else self.recycle

        if recycle:
            self._logger.info(f'Recycling worker {self.id}')
            # Shut down current worker process
            self.pipe_main.send(ShutdownTask())
            self.process.join(2)
            # Start new one
            self.process = multiprocessing.Process(
                target=self.run_worker,
                daemon=True
            )
            self.process.start()
            self._heartbeat = time.monotonic()

    def run_parent(self) -> bool:
        """
        Worker thread, running in parent process
        """
        try:
            if self.task is None:
                # we are not working on something...
                if len(self.tasks) > 0:
                    # ...but there is more work to be done
                    self.maybe_recycle()

                    try:
                        self.task = self.tasks.popleft()
                    except IndexError:
                        # another worker may have taken the last task
                        self._logger.debug(f'No more tasks to do for {self.id}')
                    else:
                        self.pipe_main.send(self.task)
                        self._logger.info(f'Dispatched task #{self.task.id} to worker {self.id} ({len(self.tasks)}  tasks left)')
                else:
                    # ...and no more work available, so we shut down this worker
                    self._logger.info(f'Worker {self.id} shutting down...')
                    self.pipe_main.send(ShutdownTask())
                    self.running = False
            else:
                if self.pipe_main.poll(0.01):
                    self._logger.debug(f'Worker {self.id} got result for task {self.task.id}...')
                    result: Result = self.pipe_main.recv()

                    if result.error:
                        self._logger.error(f'Worker {self.id} received error: {result}')
                    else:
                        self._logger.info(f'Worker {self.id} got result for task #{result.task_id}')

                    self.task.add_result(result)
                    self.finished_tasks.append(self.task)
                    self.task = None

                if self.heartbeat_main.poll():
                    hb = self.heartbeat_main.recv()
                    self._heartbeat = time.monotonic()
                    # self._logger.debug(f'Worker {self.id} got heartbeat #{hb}')

                if self._heartbeat < time.monotonic() - self.HEARTBEAT_TIMEOUT:
                    self._logger.debug(f'Worker {self.id} missed heartbeat!')
                    try:
                        self.process.join(0.2)  # try collecting process remains...
                    except:  # noqa
                        ...
                    if self.process.exitcode is not None:
                        # if we got an exitcode, the process really was killed
                        self._logger.warning(f'Worker {self.id} found dead!')
                        if self.task:  # if we were working on a task, requeue it to have it picked up by another worker...
                            self.tasks.appendleft(self.task)
                        self.running = False  # ...and shut down
        except:
            self._logger.exception(f'Unhandled error in worker {self.id}. This may result in an orphened worker process.')
            try:
                self.process.kill()
            except:
                ...

        return self.running

    def finalize(self):
        self.process.join(10)

        if self.process.exitcode is None:
            self._logger.warning(f'Worker {self.id} refused to shut down gracefully, killing it.')
            self.process.kill()
            self.process.join(2)
        self._logger.info(f'Worker {self.id} done (code {self.process.exitcode}).')

    def run_worker(self):
        """
        Entry point/main loop for the worker process
        """
        self.pid = os.getpid()
        self._shutdown = threading.Event()

        t = threading.Thread(target=self.run_worker_heartbeats, daemon=True)
        t.start()

        while self.running:
            self._logger.debug(f'Worker {self.id} ({self.pid}) waiting for tasks...')

            task = self.pipe_worker.recv()

            self._logger.debug(f'Worker {self.id} got task {task}')

            try:
                result = task.execute(self)
            except self.Shutdown:
                self.running = False
                self._shutdown.set()
            except Exception as e:
                self._logger.exception(msg := f'Error in worker process while executing {task}')
                self.pipe_worker.send(ErrorResult(msg))
            else:
                self._logger.debug(f'Worker {self.id} finished executing {task}, sending back result...')

                if result is not None:
                    self.pipe_worker.send(result)

            del task
            gc.collect()

        t.join(1.0)

    def run_worker_heartbeats(self):
        hb = 0
        while self.running:
            hb += 1
            self.heartbeat_worker.send(hb)
            self._shutdown.wait(self.HEARTBEAT_INTERVAL)


def execute_task(task: Task):
    logging.getLogger('sniffles.parallel').info(f'Working on {task}')
    return task.execute()


class SnifflesParentWorker(SnifflesWorker):
    """
    A worker class without multiprocessing, i.e. running in the main process. Used for profiling.
    """
    id: int = 0

    def __init__(self, config: Namespace, tasks: list[Task], **kwargs):  # noqa
        self.tasks = tasks
        self.task = None
        self.config = config
        self.finished_tasks: list[Task] = []
        self._log = logging.getLogger('sniffles.pworker')
        self._log.info(f'Using parent worker')

    def start(self) -> None:
        ...

    def run_parent(self) -> bool:
        count = len(self.tasks)
        for i, task in enumerate(self.tasks):
            self._log.info(f'Executing {task} ({i+1}/{count})')
            result = task.execute(self)
            task.add_result(result)
            self.finished_tasks.append(task)
        self._log.info(f'All tasks done.')

        return False

    def finalize(self):
        ...



#!/usr/bin/env python3
#
# Sniffles2
# A fast structural variant caller for long-read sequencing data
#
# Created:     26.09.2021
# Author:      Moritz Smolka
# Maintainer:  Hermann Romanek
# Contact:     sniffles@romanek.at
#
import logging
import os
import pickle
import json
import gzip
import math
from functools import cached_property
from typing import Optional

from sniffles import sv
from sniffles.config import SnifflesConfig


log = logging.getLogger(__name__)


class SNFileBase:
    """
    Base class for SNF-Files
    """
    header_length: int
    _header: Optional[dict]

    def __init__(self, config: SnifflesConfig, handle, filename=None):
        self.config = config
        self.handle = handle
        self.filename = filename
        self.blocks = {}
        self._header = None
        self._index = {}
        self.total_length = 0
        self._results = []

    @classmethod
    def open(cls, filename: str) -> 'SNFileBase':
        obj = cls(SnifflesConfig.GLOBAL, open(filename, 'rb'), filename)
        obj.read_header()
        return obj

    @property
    def index(self) -> dict:
        return self._index

    @property
    def header(self) -> dict:
        return self._header

    @cached_property
    def population(self) -> object | None:
        """
        Is this a population SNF?
        """
        return self.header.get('population', None)

    @cached_property
    def reqc(self) -> bool:
        """
        Was this file created by an old Version of sniffles we want to redo qc?
        """
        if self.config.reqc == 'auto':
            try:
                build, _, _ = self.header['config']['build'].partition('-')
            except (KeyError, AttributeError):
                log.warning(f'Unable to determine version of SNF file {self.filename} for auto-reqc')
                return True
            else:
                return build < '2.5.3'
        else:
            return self.config.reqc

    def is_open(self) -> bool:
        return self.handle is not False

    def _open(self):
        if self.handle is not False:
            self.close()
        self.handle = open(self.filename, "rb")

    def store(self, svcand):
        block_index = int(svcand.pos / self.config.snf_block_size) * self.config.snf_block_size
        if block_index not in self.blocks:
            self.blocks[block_index] = {svtype: [] for svtype in sv.TYPES}
            self.blocks[block_index]["_COVERAGE"] = {}
        if not self.config.output_rnames:
            svcand.rnames = None
        self.blocks[block_index][svcand.svtype].append(svcand)

    def serialize_block(self, block_id):
        return pickle.dumps(self.blocks[block_id])

    def unserialize_block(self, data: bytes):
        return pickle.loads(data)

    def write_and_index(self):
        if not self.is_open():
            self._open()
        offset = 0
        for block_id in sorted(self.blocks):
            data = gzip.compress(self.serialize_block(block_id))
            self.handle.write(data)
            data_len = len(data)
            self._index[block_id] = (offset, data_len)
            offset += data_len
            self.total_length += data_len
        if self.config.combine_close_handles:
            self.close()

    def read_header(self):
        if not self.is_open():
            self._open()
        try:
            header_text = self.handle.readline()
            self.header_length = len(header_text)
            self._header = json.loads(header_text.strip())
        except Exception as e:
            print(f"Error when reading SNF header from '{self.filename}': {e}. The file may not be a valid .snf file or could have been corrupted.")
            raise e
        self._index = self._header["index"]
        if self.config.combine_close_handles:
            self.close()

    def read_blocks(self, contig, block_index):
        if not self.is_open():
            self._open()
        block_index = str(block_index)
        if contig not in self.index:
            if self.config.combine_close_handles:
                self.close()
            return None

        if block_index not in self.index[contig]:
            if self.config.combine_close_handles:
                self.close()
            return None

        blocks = []
        for block_data_start, block_data_length in self.index[contig][block_index]:
            try:
                self.handle.seek(self.header_length + block_data_start)
                data = gzip.decompress(self.handle.read(block_data_length))
                blocks.append(self.unserialize_block(data))
            except Exception as e:
                print(f"Error when reading block '{contig}.{block_index}' from '{self.filename}': {e}. The file may not be a valid .snf file or could have been corrupted.")
                if self.config.combine_close_handles:
                    self.close()
                raise e
        if self.config.combine_close_handles:
            self.close()
        return blocks

    def get_index(self):
        return self.index

    def get_total_length(self):
        return self.total_length

    def add_result(self, result):
        if result.has_snf:
            self._results.append(result)

    def _calculate_contig_coverages(self, contigs: list[str]) -> dict[str, float]:
        contig_coverages: dict[str, list[float]] = {
            c: [] for c in contigs
        }

        for r in self._results:
            contig_coverages[r.contig].append(r.coverage_average_total)

        res = {}
        for contig in contig_coverages:
            res[contig] = sum(contig_coverages[contig]) / len(contig_coverages[contig]) if len(contig_coverages[contig]) > 0 else 0

        return res

    def _create_header(self, config: SnifflesConfig, main_index: dict, snf_candidate_count: int) -> dict:
        """
        Generate the header for this SNF file.
        """
        return {"config": config.__dict__, "index": main_index, "snf_candidate_count": snf_candidate_count}

    def write_results(self, config: SnifflesConfig, contigs: list[str]) -> int:
        """
        Writes all added results (regional temporary .snf files) to this file. Returns SNF candidate count
        """
        main_index = {}
        offset = 0
        snf_candidate_count = sum(r.snf_candidate_count for r in self._results)
        parts_sorted = sorted(self._results, key=lambda r: r.task_id)
        for part in parts_sorted:
            part_contig = part.contig
            if part_contig not in main_index:
                main_index[part_contig] = {}
            for block, (part_block_start, part_block_len) in part.snf_index.items():
                if block not in main_index[part_contig]:
                    main_index[part_contig][block] = []
                main_index[part_contig][block].append((part_block_start + offset, part_block_len))
            offset += part.snf_total_length

        config.contig_coverages = self._calculate_contig_coverages(contigs)
        header = self._create_header(config, main_index, snf_candidate_count)
        header_json = json.dumps(header, default=lambda obj: "<Unstored_Object>") + "\n"
        self.handle.write(header_json.encode())

        for part in parts_sorted:
            with open(part.snf_filename, "rb") as part_handle:
                part_data = part_handle.read()
                part_handle.close()
            self.handle.write(part_data)
            os.remove(part.snf_filename)

        return snf_candidate_count

    def close(self) -> None:
        if self.handle is not False:
            self.handle.close()
            self.handle = False

    def get_all_blocks(self, contig: str) -> dict[str, dict]:
        """
        Fetches all blocks for given contig.
        """
        blocks = {}
        if contig in self.index:
            for block_start in self.index[contig].keys():
                blocks[block_start] = self.read_blocks(contig, block_start)[0]
        return blocks

    def get_full_coverage(self, contig: str):
        coverage = {}
        for b in self.get_all_blocks(contig).values():
            coverage.update(b['_COVERAGE'])
        return coverage


class SNFile(SNFileBase):
    def annotate_block_coverages(self, lead_provider, resolution=500):
        config = self.config
        start_bin = lead_provider.covrtab_min_bin
        end_bin = int(lead_provider.end / config.coverage_binsize) * config.coverage_binsize
        coverage_fwd = 0
        coverage_rev = 0

        coverage_sum = 0
        bin_count = 0

        coverage_binsize_combine = self.config.coverage_binsize_combine
        snf_block_size = config.snf_block_size

        for bin in range(start_bin, end_bin + config.coverage_binsize, config.coverage_binsize):
            if bin in lead_provider.covrtab_fwd:
                coverage_fwd += lead_provider.covrtab_fwd[bin]

            if bin in lead_provider.covrtab_rev:
                coverage_rev += lead_provider.covrtab_rev[bin]

            coverage_sum += coverage_fwd + coverage_rev
            bin_count += 1

            if bin % coverage_binsize_combine == 0:
                block_index = int(bin / snf_block_size) * snf_block_size

                coverage_total_curr = math.ceil(coverage_sum / float(bin_count))
                if coverage_total_curr > 0:
                    if block_index not in self.blocks:
                        self.blocks[block_index] = {svtype: [] for svtype in sv.TYPES}
                        self.blocks[block_index]["_COVERAGE"] = {}

                    self.blocks[block_index]["_COVERAGE"][bin] = coverage_total_curr

                coverage_sum = 0
                bin_count = 0


class RemoteIndexSNFile(SNFile):
    """
    An SNFile, but its header data will be provided by a remote source.
    """

    @property
    def index(self) -> dict:
        if self._header is None:
            self.read_header()

        return super().index

    @property
    def header(self) -> dict:
        if self._header is None:
            self.read_header()

        return super().header



#!/usr/bin/env python3
#
# Sniffles2
# A fast structural variant caller for long-read sequencing data
#
# Created:     27.08.2021
# Author:      Moritz Smolka
# Maintainer:  Hermann Romanek
# Contact:     sniffles@romanek.at
#

from sniffles import util
from sniffles import consensus
from sniffles.config import SnifflesConfig
from sniffles.sv import SVCall
import math


def annotate_sv(svcall: SVCall, config):
    if config.phase:
        phase = phase_sv(svcall, config)
    else:
        phase = (None, None)

    genotype_sv(svcall, config, phase)

    if svcall.svtype == "INS" and not config.symbolic:
        merged_leads = [lead for lead in svcall.postprocess.cluster.leads if lead.seq is not None]

        if len(merged_leads):
            best_lead = merged_leads[0]
            best_index = 0
            best_diff = abs(len(best_lead.seq) - svcall.svlen) + abs(best_lead.ref_start - svcall.pos) * 1.5
            for i, ld in enumerate(merged_leads):
                if i == 0:
                    continue
                curr_diff = abs(len(ld.seq) - svcall.svlen) + abs(ld.ref_start - svcall.pos) * 1.5
                if curr_diff < best_diff:
                    best_lead = ld
                    best_index = i
                    best_diff = curr_diff

            merged_leads.pop(best_index)
            # merged_leads_new=list()

            # for lead in merged_leads:
            #    curr_lend_iff=abs(len(ld.seq) - len(best_lead.seq)) + abs(ld.ref_start - best_lead.ref_start)
            #    #if curr_len_diff < 25:
            #    merged_leads_new.append(lead)
            # merged_leads=merged_leads_new

            if len(merged_leads) >= config.consensus_min_reads and not config.no_consensus:
                kmer_len = config.consensus_kmer_len
                skip = config.consensus_kmer_skip_base + int(len(best_lead.seq)*config.consensus_kmer_skip_seqlen_mult)
                skip_repetitive = skip

                svcall.alt = consensus.novel_from_reads(best_lead, merged_leads, klen=kmer_len, skip=skip,
                                                        skip_repetitive=skip_repetitive)
            else:
                svcall.alt = best_lead.seq


def add_request(svcall, field, pos, requests_for_coverage, config):
    bin_index = int(pos / config.coverage_binsize) * config.coverage_binsize
    if bin_index not in requests_for_coverage:
        requests_for_coverage[bin_index] = []
    requests_for_coverage[bin_index].append((svcall, field))


def coverage(calls, lead_provider, config):
    requests_for_coverage = coverage_build_requests(calls, config)
    return coverage_fulfill(requests_for_coverage, lead_provider, config)


def coverage_build_requests(calls, config: SnifflesConfig):
    """
    Updated: changed coverage calculation to be inside the SV for INV/DEL/DUP, outside (like it is now) for INS/BND?
    BIN: ---
    INS/BND    |-------------V---------------|
              U---       S--- E---            D---
                            C---
    DUP/DEL    |------------===============------------|
    INV       U---          S---       E---            D---
                                  C---
    """
    requests_for_coverage = {}
    for svcall in calls:
        start = svcall.pos
        if svcall.svtype == "INS":
            end = start + 1
        else:
            end = svcall.pos + abs(svcall.svlen)
        if svcall.svtype in ["INS", "BND"]:
            add_request(svcall, "coverage_start", start - config.coverage_binsize, requests_for_coverage, config)
            add_request(svcall, "coverage_center", int((start + end - config.coverage_binsize) / 2),
                        requests_for_coverage, config)
            add_request(svcall, "coverage_end", end + config.coverage_binsize, requests_for_coverage, config)
            add_request(svcall, "coverage_upstream", start - config.coverage_binsize * config.coverage_updown_bins,
                        requests_for_coverage, config)
            add_request(svcall, "coverage_downstream", end + config.coverage_binsize * config.coverage_updown_bins,
                        requests_for_coverage, config)
        else:
            add_request(svcall, "coverage_start", start, requests_for_coverage, config)
            add_request(svcall, "coverage_center", int((start + end - config.coverage_binsize) / 2),
                        requests_for_coverage, config)
            add_request(svcall, "coverage_end", end - config.coverage_binsize, requests_for_coverage, config)
            add_request(svcall, "coverage_upstream", start - config.coverage_binsize * config.coverage_updown_bins,
                        requests_for_coverage, config)
            add_request(svcall, "coverage_downstream", end + config.coverage_binsize * config.coverage_updown_bins,
                        requests_for_coverage, config)

    return requests_for_coverage


def coverage_fulfill(requests_for_coverage, lead_provider, config: SnifflesConfig):
    if len(requests_for_coverage) == 0:
        return -1, -1

    start_bin = lead_provider.covrtab_min_bin
    end_bin = int(lead_provider.end / config.coverage_binsize) * config.coverage_binsize
    coverage_fwd = 0
    coverage_rev = 0
    coverage_fwd_total = 0
    coverage_rev_total = 0
    n = 0

    for bin_index in range(start_bin, end_bin + config.coverage_binsize, config.coverage_binsize):
        n += 1

        if bin_index in lead_provider.covrtab_fwd:
            coverage_fwd += lead_provider.covrtab_fwd[bin_index]

        if bin_index in lead_provider.covrtab_rev:
            coverage_rev += lead_provider.covrtab_rev[bin_index]

        if bin_index in requests_for_coverage:
            coverage_total_curr = coverage_fwd + coverage_rev
            for svcall, field in requests_for_coverage[bin_index]:
                setattr(svcall, field, coverage_total_curr)

        coverage_fwd_total += coverage_fwd
        coverage_rev_total += coverage_rev

    average_coverage_fwd = coverage_fwd_total / float(n) if n > 0 else 0
    average_coverage_rev = coverage_rev_total / float(n) if n > 0 else 0
    return average_coverage_fwd, average_coverage_rev


def qc_sv_support(svcall, coverage_global, config):
    if config.mosaic:
        return True
    if config.minsupport == "auto":
        if not qc_support_auto(svcall, coverage_global, config):
            svcall.filter = "SUPPORT_MIN"
            return False
    else:
        if not qc_support_const(svcall, config):
            svcall.filter = "SUPPORT_MIN"
            return False
    return True


def rescale_support(svcall, config: SnifflesConfig):
    if svcall.svtype != "INS" or svcall.svlen < config.long_ins_length:
        return svcall.support
    else:
        base = svcall.support
        scale_factor = config.long_ins_rescale_mult * (float(svcall.svlen) / config.long_ins_length)
        return round(base * (config.long_ins_rescale_base + scale_factor))


def qc_support_auto(svcall, coverage_global, config):
    support = rescale_support(svcall, config)
    # if svcall.svtype=="INS":
    #    coverage_list=[svcall.coverage_center]
    # else:
    coverage_list = [each_coverage for each_coverage in [svcall.coverage_upstream, svcall.coverage_downstream] if
                     each_coverage != 0]
    if len(coverage_list) == 0:
        coverage_list = [each_coverage for each_coverage in [svcall.coverage_start, svcall.coverage_center,
                                                             svcall.coverage_end] if each_coverage != 0]
    if len(coverage_list) == 0:
        coverage_regional = coverage_global
    else:
        coverage_regional = round(sum(coverage_list) / len(coverage_list))
        if coverage_regional == 0:
            coverage_regional = coverage_global
    coverage_global_weight = (1.0 - config.minsupport_auto_regional_coverage_weight)
    coverage = (coverage_regional * config.minsupport_auto_regional_coverage_weight +
                coverage_global * coverage_global_weight)
    min_support = round(config.minsupport_auto_base + config.minsupport_auto_mult * coverage)
    return support >= min_support
    # return True


def qc_support_const(svcall, config):
    # svcall.set_info("MINSUPPORT",config.minsupport)
    return svcall.support >= config.minsupport


def qc_sv(svcall: SVCall, config: SnifflesConfig):
    af = svcall.get_info("VAF")
    af = af if af is not None else 0
    sv_is_mosaic = af <= config.mosaic_af_max

    if config.qc_stdev:
        stdev_pos = svcall.get_info("STDEV_POS")
        if stdev_pos > config.qc_stdev_abs_max:
            svcall.filter = "STDEV_POS"
            return False
        if svcall.svtype != "BND" and stdev_pos / abs(svcall.svlen) > 2.0:
            svcall.filter = "STDEV_POS"
            return False

        stdev_len = svcall.get_info("STDEV_LEN")
        if stdev_len is not None:
            if svcall.svtype != "BND" and stdev_len / abs(svcall.svlen) > 1.0:
                svcall.filter = "STDEV_LEN"
                return False
            if stdev_len > config.qc_stdev_abs_max:
                svcall.filter = "STDEV_LEN"
                return False

    if config.mosaic and sv_is_mosaic:
        if svcall.support < config.mosaic_min_reads:
            svcall.filter = "SUPPORT_MIN"
            return False

    if abs(svcall.svlen) < config.minsvlen:
        svcall.filter = "SVLEN_MIN"
        return False

    if svcall.svtype == "BND":
        if config.qc_bnd_filter_strand and len(set(lead.strand for lead in svcall.postprocess.cluster.leads)) < 2:
            svcall.filter = "STRAND_BND"
            return False
    elif not (config.mosaic and sv_is_mosaic) and config.qc_strand:
        is_long_ins = (svcall.svtype == "INS" and svcall.svlen >= config.long_ins_length)
        if not is_long_ins and len(set(leads.strand for leads in svcall.postprocess.cluster.leads)) < 2:
            svcall.filter = "STRAND"
            return False
    elif (config.mosaic and sv_is_mosaic) and config.mosaic_qc_strand:
        is_long_ins = (svcall.svtype == "INS" and svcall.svlen >= config.long_ins_length)
        if (not is_long_ins and len(set(leads.strand for leads in svcall.postprocess.cluster.leads)) < 2
                and svcall.support >= config.mosaic_use_strand_thresholds):
            svcall.filter = "STRAND_MOSAIC"
            return False

    if config.mosaic and sv_is_mosaic:
        if (svcall.svtype == "INV" or svcall.svtype == "DUP") and svcall.svlen < config.mosaic_qc_invdup_min_length:
            svcall.filter = "SVLEN_MIN_MOSAIC"
            return False

    if svcall.coverage_center < config.qc_coverage and svcall.svtype not in ["DEL", "INS"]:
        if (svcall.svtype == "INV" and svcall.svlen) > config.long_inv_length and not (config.mosaic and sv_is_mosaic):
            pass
        else:
            svcall.filter = "COV_MIN"
            return False

    upstream_downstream_max_coverage_diff = 0.7  # 70%
    upstream_downstream_diff = 0.5  # add to config
    if (svcall.svtype == "DEL" and config.long_del_length != -1 and abs(svcall.svlen) >= config.long_del_length and
            not config.mosaic):
        scaled_long_del_coverage = config.long_del_coverage/2.0   # 0.66/2 = 0.33
        if svcall.coverage_center > (svcall.coverage_upstream + svcall.coverage_downstream) * scaled_long_del_coverage:
            # check if slopped coverage, that often happens over large spans
            if svcall.coverage_upstream > svcall.coverage_center > svcall.coverage_downstream:
                if svcall.coverage_downstream/svcall.coverage_upstream < upstream_downstream_max_coverage_diff:
                    svcall.filter = "COV_CHANGE_DEL"
                    return False
            elif svcall.coverage_upstream < svcall.coverage_center < svcall.coverage_downstream:
                if svcall.coverage_upstream/svcall.coverage_downstream < upstream_downstream_max_coverage_diff:
                    svcall.filter = "COV_CHANGE_DEL"
                    return False
            else:
                pass
        if svcall.coverage_upstream > svcall.coverage_downstream:
            if (upstream_downstream_diff > svcall.coverage_downstream/svcall.coverage_upstream or
                    svcall.coverage_center > svcall.coverage_downstream):
                svcall.filter = "COV_CHANGE_DEL"
                return False
        elif svcall.coverage_upstream < svcall.coverage_downstream:
            if (upstream_downstream_diff > svcall.coverage_upstream/svcall.coverage_downstream or
                    svcall.coverage_upstream < svcall.coverage_center):
                svcall.filter = "COV_CHANGE_DEL"
                return False
        else:
            pass
    elif (svcall.svtype == "DUP" and config.long_dup_length != -1 and abs(svcall.svlen) >= config.long_dup_length and
          not config.mosaic):
        scaled_long_dup_coverage = config.long_dup_coverage / 2.0
        if svcall.coverage_center < (svcall.coverage_upstream + svcall.coverage_downstream) * scaled_long_dup_coverage:
            # check if slopped coverage, that often happens over large spans
            if svcall.coverage_upstream > svcall.coverage_center > svcall.coverage_downstream:
                if svcall.coverage_downstream/svcall.coverage_upstream < upstream_downstream_max_coverage_diff:
                    svcall.filter = "COV_CHANGE_DUP"
                    return False
            elif svcall.coverage_upstream < svcall.coverage_center < svcall.coverage_downstream:
                if svcall.coverage_upstream/svcall.coverage_downstream < upstream_downstream_max_coverage_diff:
                    svcall.filter = "COV_CHANGE_DUP"
                    return False
            else:
                pass
            if svcall.coverage_upstream > svcall.coverage_downstream:
                if (upstream_downstream_diff > svcall.coverage_downstream / svcall.coverage_upstream or
                        svcall.coverage_center < svcall.coverage_downstream):
                    svcall.filter = "COV_CHANGE_DUP"
                    return False
            elif svcall.coverage_upstream < svcall.coverage_downstream:
                if (upstream_downstream_diff > svcall.coverage_upstream / svcall.coverage_downstream or
                        svcall.coverage_upstream > svcall.coverage_center):
                    svcall.filter = "COV_CHANGE_DUP"
                    return False
            else:
                pass
    elif svcall.svtype == "INS" and (svcall.coverage_upstream < config.qc_coverage or
                                     svcall.coverage_downstream < config.qc_coverage):
        svcall.filter = "COV_CHANGE_INS"
        return False

    qc_coverage_max_change_frac = config.qc_coverage_max_change_frac
    if config.mosaic and sv_is_mosaic:
        qc_coverage_max_change_frac = config.mosaic_qc_coverage_max_change_frac
    if qc_coverage_max_change_frac != -1.0:
        if svcall.coverage_upstream != 0:
            u = float(svcall.coverage_upstream)
        else:
            u = 1.0

        if svcall.coverage_start != 0:
            s = float(svcall.coverage_start)
        else:
            s = 1.0

        if svcall.coverage_center != 0:
            c = float(svcall.coverage_center)
        else:
            c = 1.0

        if svcall.coverage_end != 0:
            e = float(svcall.coverage_end)
        else:
            e = 1.0

        if svcall.coverage_downstream != 0:
            d = float(svcall.coverage_downstream)
        else:
            d = 1.0

        if abs(u - s) / max(u, s) > qc_coverage_max_change_frac:
            svcall.filter = "COV_CHANGE_FRAC_US"
            return False

        if abs(s - c) / max(s, c) > qc_coverage_max_change_frac:
            svcall.filter = "COV_CHANGE_FRAC_SC"
            return False

        if abs(c - e) / max(c, e) > qc_coverage_max_change_frac:
            svcall.filter = "COV_CHANGE_FRAC_CE"
            return False

        if abs(e - d) / max(e, d) > qc_coverage_max_change_frac:
            svcall.filter = "COV_CHANGE_FRAC_ED"
            return False

    return True


def qc_sv_post_annotate(svcall: SVCall, config: SnifflesConfig):
    af = svcall.get_info("VAF")
    af = af if af is not None else 0
    sv_is_mosaic = af <= config.mosaic_af_max

    if (svcall.coverage_center < config.qc_coverage and
            (len(svcall.genotypes) == 0 or (svcall.genotypes[0][0] != "." and
                                            svcall.genotypes[0][0] + svcall.genotypes[0][1] < 2))):
        svcall.filter = "COV_MIN_GT"
        return False

    qc_nm = config.qc_nm
    qc_nm_threshold = config.qc_nm_threshold * config.qc_nm_mult
    if config.mosaic and sv_is_mosaic:
        qc_nm = config.mosaic_qc_nm
        qc_nm_threshold = config.qc_nm_threshold * config.qc_nm_mult
    if qc_nm and svcall.nm > qc_nm_threshold and (len(svcall.genotypes) == 0 or svcall.genotypes[0][1] == 0):
        svcall.filter = "ALN_NM"
        return False

    if config.mosaic:
        if sv_is_mosaic and (af < config.mosaic_af_min or af > config.mosaic_af_max):
            svcall.filter = "MOSAIC_VAF"
            return False
        elif not sv_is_mosaic and not config.mosaic_include_germline:
            svcall.filter = "NOT_MOSAIC_VAF"
            return False

    return True


def binomial_coef(n, k):
    return math.factorial(n) / (math.factorial(k) * math.factorial(n - k))


def genotype_sv(svcall: SVCall, config, phase: tuple | None = None):
    from sniffles.genotyping import GENOTYPER_BY_TYPE, Genotyper

    GENOTYPER_BY_TYPE.get(svcall.svtype, Genotyper)(svcall, config, phase).calculate()


def phase_sv(svcall, config):
    reads_phases = {lead.read_id[0]: (lead.read_id[1], lead.read_id[2]) for lead in svcall.postprocess.cluster.leads}
    hp_list = util.most_common(hp for hp, ps in reads_phases.values())
    ps_list = util.most_common(ps for hp, ps in reads_phases.values())

    hp_support, hp = hp_list[0]
    ps_support, ps = ps_list[0]
    if hp is None:
        hp = "NULL"
    if ps is None:
        ps = "NULL"

    other_hp_support = sum(other_supp for other_supp, other_hp in hp_list if other_hp != hp and other_hp != "NULL")
    other_ps_support = sum(other_supp for other_supp, other_ps in ps_list if other_ps != ps and other_ps != "NULL")

    hp_filter = "FAIL"
    if (float(other_hp_support) / (hp_support + other_hp_support) < config.phase_conflict_threshold and hp != "NULL" and
            hp_support > 0):
        hp_filter = "PASS"

    ps_filter = "FAIL"
    if (float(other_ps_support) / (ps_support + other_ps_support) < config.phase_conflict_threshold and ps != "NULL" and
            ps_support > 0):
        ps_filter = "PASS"

    svcall.set_info("PHASE", f"{hp},{ps},{hp_support},{ps_support},{hp_filter},{ps_filter}")
    return (config.phase_identifiers.index(hp) if hp in config.phase_identifiers else None if hp_filter == "PASS" else
            None, ps if "PASS" == ps_filter else None)



#!/usr/bin/env python3
#
# Sniffles2
# A fast structural variant caller for long-read sequencing data
#
# Created:     20.09.2024
# Author:      Hermann Romanek
# Maintainer:  Hermann Romanek
# Contact:     sniffles@romanek.at
#
"""
Genotyping
"""
import math
from dataclasses import dataclass
from typing import Any

from sniffles.postprocessing import rescale_support
from sniffles.sv import SVCall


class UnknownGenotypeError(Exception):
    """
    Unable to determine genotype
    """


def binomial_probability(k, n, p):
    try:
        # Binomial coef cancels out for likelihood ratios
        return (p ** k) * ((1.0 - p) ** (n - k))
    except OverflowError:
        return 1.0


def likelihood_ratio(q1, q2):
    if q1 / q2 > 0:
        try:
            return math.log(q1 / q2, 10)
        except ValueError:
            return 0
    else:
        return 0


class UnknownGenotype:
    ...


@dataclass
class Genotype:
    a: int
    b: int
    qual: int  # GQ, 0-60
    dr: int
    dv: int
    phase: Any

    UNKNOWN = UnknownGenotype()



class Genotyper:
    """
    Generic genotyping
    """
    _support: int
    _coverage: float

    def __init__(self, svcall: SVCall, config, phase: tuple | None):
        self.svcall = svcall
        self.config = config
        self.phase = phase if phase is not None else self._get_phase()

    def _get_phase(self) -> tuple | None:
        """
        Get phasing information from the genotyped SV, or None if it can't be determined.
        """
        try:
            return self.svcall.genotypes[0][5]
        except (KeyError, IndexError):
            return None

    def _calculate_support(self) -> int:
        return self.svcall.support

    def _calculate_coverage(self, support: int) -> int:
        return self._get_coverage_from_list()

    def _calculate_af(self, support: int, coverage: int) -> float:
        """
        Calculate allele fraction
        """
        return support / float(coverage)

    def _get_coverage_from_list(self, coverage_list: list = None) -> int:
        """
        Coverage here is NOT the same as in IGV, but the number of reads spanning the SV
        """
        svcall = self.svcall
        if coverage_list is None:
            coverage_list = [svcall.coverage_start, svcall.coverage_center, svcall.coverage_end]

        coverage_list = [each_coverage for each_coverage in coverage_list if each_coverage != 0]

        if len(coverage_list) > 0:
            if None in coverage_list:
                new_coverage_list = [cov_value for cov_value in coverage_list if cov_value is not None]
                if len(new_coverage_list) > 0:
                    return round(sum(new_coverage_list) / len(new_coverage_list))
                else:
                    raise UnknownGenotypeError()
            else:
                return round(sum(coverage_list) / len(coverage_list))
        else:
            raise UnknownGenotypeError()

    def _filter_by_z_score(self, z_score: float) -> bool:
        """
        Should this call be filtered due to z score?
        """
        return z_score < self.config.genotype_min_z_score and not self.config.mosaic

    def calculate(self):
        config = self.config
        normalization_target = 250
        hom_ref_p = config.genotype_error
        het_p = (1.0 / config.genotype_ploidy)  # - config.genotype_error
        hom_var_p = 1.0 - config.genotype_error
        svcall = self.svcall

        support = self._calculate_support()
        try:
            coverage = self._calculate_coverage(support)
        except UnknownGenotypeError:
            return

        if support > coverage:
            coverage = support

        af = self._calculate_af(support, coverage)

        genotype_p = [((0, 0), hom_ref_p),
                      ((0, 1), het_p),
                      ((1, 1), hom_var_p)]

        max_lead = max(support, coverage)
        if max_lead > normalization_target:
            norm = normalization_target / float(max_lead)
            normalized_support = round(support * norm)
            normalized_coverage = round(coverage * norm)
        else:
            normalized_support = support
            normalized_coverage = coverage

        genotype_likelihoods = []
        for gt, p in genotype_p:
            q = binomial_probability(normalized_support, normalized_coverage, p)
            genotype_likelihoods.append((gt, q))
        genotype_likelihoods.sort(key=lambda k: k[1], reverse=True)

        sum_likelihoods = sum(q for gt, q in genotype_likelihoods)
        normalized_likelihoods = [(gt, (q / sum_likelihoods)) for gt, q in genotype_likelihoods]

        gt1, q1 = normalized_likelihoods[0]
        gt2, q2 = normalized_likelihoods[1]
        qz = [q for gt, q in normalized_likelihoods if gt == (0, 0)][0]
        genotype_z_score = min(60, int((-10) * likelihood_ratio(qz, q1)))
        genotype_quality = min(60, int((-10) * likelihood_ratio(q2, q1)))

        if svcall.filter == "PASS" and self._filter_by_z_score(genotype_z_score):
            svcall.filter = "GT"

        a, b = gt1
        svcall.genotypes[0] = (a, b, genotype_quality, coverage - support, support, self.phase)
        svcall.set_info("VAF", af)


class InsertionGenotyper(Genotyper):
    def _calculate_support(self):
        """
        Rescale support for long insertions. This skews our support way higher than the number of reads we have!
        """
        return rescale_support(self.svcall, self.config)

    def _calculate_coverage(self, coverage_list: list = None) -> float:
        return self._get_coverage_from_list([self.svcall.coverage_center])

    def _filter_by_z_score(self, z_score: float) -> bool:
        """
        When detecting large insertions, we allow for below threshold z scores
        """
        flt = super()._filter_by_z_score(z_score)
        if flt and self.svcall.svlen >= self.config.long_ins_length and self.config.detect_large_ins:
            return False
        return flt


class DuplicationGenotyper(Genotyper):
    def _calculate_coverage(self, support: int) -> float:
        # Experimental other coverage calculation?
        # if False and svcall.coverage_start != 0 and svcall.coverage_end != 0:
        #     if svcall.coverage_start > svcall.coverage_end:
        #         coverage_list = [svcall.coverage_end]
        #     else:
        #         coverage_list = [svcall.coverage_start]
        svcall = self.svcall
        return self._get_coverage_from_list([svcall.coverage_start, svcall.coverage_end]) + round(support * 0.75)


class InversionGenotyper(Genotyper):

    def _calculate_coverage(self, support: int) -> int:
        svcall = self.svcall
        # check event length, for whole chromosome event do something different
        return self._get_coverage_from_list([svcall.coverage_upstream, svcall.coverage_downstream]) + round(support * 0.5)


class DeletionGenotyper(Genotyper):

    def _calculate_coverage(self, support: int) -> int:
        svcall = self.svcall
        if support_sa := svcall.get_info('SUPPORT_SA'):
            return self._get_coverage_from_list([svcall.coverage_start + support_sa, svcall.coverage_center + support_sa, svcall.coverage_end + support_sa])
        else:
            return super()._calculate_coverage(support)


GENOTYPER_BY_TYPE = {
    'INS': InsertionGenotyper,
    'DEL': DeletionGenotyper,
    'DUP': DuplicationGenotyper,
    'INV': InversionGenotyper,
}



#!/usr/bin/env python3
#
# Sniffles2
# A fast structural variant caller for long-read sequencing data
#
# Created:     05.12.2023
# Author:      Hermann Romanek
# Maintainer:  Hermann Romanek
# Contact:     sniffles@romanek.at
#
import logging
import os
import pickle

from sniffles.config import SnifflesConfig
from sniffles.snf import SNFile
from sniffles.sv import SVCall
from sniffles.vcf import VCF


log = logging.getLogger(__name__)


class Result:
    """
    A generic result of a sniffles task executed by a worker process.
    """
    processed_read_count: int
    task_id: int
    run_id: str
    contig: str
    processed_read_count: int
    svcalls: list[SVCall]
    svcount: int
    error: bool = False

    def __init__(self, task: 'Task', svcalls: list[SVCall], candidates_processed: int):
        self.task_id = task.id
        self.run_id = task.config.run_id
        self.contig = task.contig
        self.processed_read_count = candidates_processed
        self.svcount = len(svcalls)
        self.store_calls(svcalls)

    def store_calls(self, svcalls: list[SVCall]) -> None:
        """
        Store the given calls to this result
        """
        self.svcalls = svcalls

    def emit(self, vcf_out: VCF = None, **kwargs) -> int:
        """
        Emit this result to a file. Returns the number of records written.
        """
        if vcf_out is not None:
            calls = self.svcalls
            if calls:
                for call in calls:
                    vcf_out.write_call(call)
                log.debug(f"Wrote {len(calls)} calls from {self} to VCF.")
            else:
                log.debug(f'No calls for {self}')
            return len(calls)
        else:
            log.debug(f'No vcf output file specified.')
            return 0


class CallResult(Result):
    coverage_average_total: float
    has_snf = False
    snf_filename = None
    snf_index = None
    snf_total_length = None
    snf_candidate_count = None

    def emit(self, **kwargs) -> int:
        res = super().emit(**kwargs)
        if snf_out := kwargs.get('snf_out'):
            snf_out.add_result(self)
        return res


class GenotypeResult(Result):
    """
    Result of a genotyping run
    """

    def emit(self, vcf_out: VCF = None, snf_out: SNFile = None, **kwargs) -> int:
        if vcf_out is not None:
            genotype_lineindex_order = kwargs['genotype_lineindex_order']
            genotype_lineindex_svcalls_returned = {}

            for svcall in self.svcalls:
                genotype_lineindex_svcalls_returned[svcall.raw_vcf_line_index] = svcall

            for lineindex in genotype_lineindex_order:
                if lineindex in genotype_lineindex_svcalls_returned:
                    vcf_out.rewrite_genotype(genotype_lineindex_svcalls_returned[lineindex])

        return len(self.svcalls)


class CombineResult(Result):
    """
    Result of a combine run for one task, simple variant with calls in memory. Must be pickleable.
    """
    def store_calls(self, svcalls: list[SVCall]) -> None:
        from sniffles.config import SnifflesConfig
        if SnifflesConfig.GLOBAL.sort:
            svcalls = sorted(svcalls, key=lambda call: call.pos)

        try:
            self.svcalls.extend(svcalls)
        except AttributeError:
            self.svcalls = list(svcalls)

    def finalize(self):
        if SnifflesConfig.GLOBAL.sort:
            self.svcalls.sort(key=lambda call: call.pos)

    def emit(self, vcf_out: VCF = None, **kwargs) -> int:
        res = super().emit(vcf_out=vcf_out, **kwargs)
        from sniffles.config import SnifflesConfig
        if (config := SnifflesConfig.GLOBAL).dev_population_snf:
            from sniffles.snfp import PopulationSNF
            PopulationSNF.create(config, self.svcalls)

        return res

    def __str__(self):
        return f'CombineResult #{self.task_id}'


class CombineResultTmpFile(CombineResult):
    """
    Result of a combine run, with calls in a temporary file instead of memory
    """
    _highest_position_call: int = -1  # maximum position of last emitted call, for sorting
    unsorted: bool = False
    _initialized: bool = False

    @property
    def tmpfile_name(self) -> str:
        return os.path.join(SnifflesConfig.GLOBAL.tmp_dir, f'result-{self.run_id}-{self.task_id:04}.part.vcf')

    @property
    def tmpfile_unsorted(self) -> str:
        """
        Name of this task/results temporary file for unsorted calls
        """
        return os.path.join(SnifflesConfig.GLOBAL.tmp_dir, f'result-{self.run_id}-{self.task_id:04}-unsorted.part.vcf')

    def store_calls(self, svcalls):
        from sniffles.config import SnifflesConfig
        offset = 0

        if not self._initialized:
            if os.path.exists(self.tmpfile_name):
                self.cleanup()
            self._initialized = True

        if SnifflesConfig.GLOBAL.sort and svcalls:
            svcalls = list(sorted(svcalls, key=lambda call: call.pos))

            while offset < len(svcalls) and svcalls[offset].pos < self._highest_position_call:
                log.debug(f'Unsorted call detected: {self._highest_position_call} > {svcalls[0]}')
                offset += 1

            if offset > 0:  # unsorted calls
                self.unsorted = True
                with open(self.tmpfile_unsorted, 'a') as f:
                    vcf = VCF(SnifflesConfig.GLOBAL, f)
                    for call in svcalls[:offset]:
                        vcf.write_call(call)

            self._highest_position_call = svcalls[-1].pos

        log.debug(f'Storing {len(svcalls)} calls for task {self.task_id} to {self.tmpfile_name}')

        with open(self.tmpfile_name, 'a') as f:
            vcf = VCF(SnifflesConfig.GLOBAL, f)
            vcf.open_reference(generate_index=False)
            for call in svcalls[offset:]:
                vcf.write_call(call)

    def finalize(self):
        ...

    def _emit_sorted(self) -> int:
        # contig, position, _ = line.split('\t', 2)
        # position = int(position)
        # while position > self._unsorted_calls[-1].pos:
        #     vcf_out.write_call(self._unsorted_calls.pop())
        #     n += 1
        return 0

    def emit(self, vcf_out: VCF = None, **kwargs) -> int:
        n = 0
        with open(self.tmpfile_name, 'r') as f:
            for line in f:
                vcf_out.handle.write(line)
                n += 1

        vcf_out.call_count += n
        self.cleanup()
        return n

    def cleanup(self):
        try:
            os.unlink(self.tmpfile_name)
        except FileNotFoundError:
            ...


class CombineResultTmpFilePopulationSNF(CombineResultTmpFile):
    """
    Combine result with temporary files plus population SNF creation
    """
    @property
    def snf_filename(self) -> str:
        return os.path.join(SnifflesConfig.GLOBAL.tmp_dir, f'result-{self.run_id}-{self.task_id:04}.part.snf')

    has_snf = True
    snf_index: dict
    snf_total_length: int
    snf_candidate_count: int = 0
    svcalls: list = None

    def store_calls(self, svcalls: list[SVCall]) -> None:
        super().store_calls(svcalls)

        if self.svcalls is None:
            self.svcalls = []
        self.svcalls.extend(svcalls)

    def finalize(self):
        with open(self.snf_filename, 'wb') as handle:
            from sniffles.snfp import PopulationSNF
            snfp = PopulationSNF(SnifflesConfig.GLOBAL, handle)
            c = 0
            for call in self.svcalls:
                c += 1 if snfp.store(call) else 0
            snfp.write_and_index()

        self.snf_index = snfp.get_index()
        self.snf_total_length = snfp.get_total_length()
        self.snf_candidate_count = c

    def emit(self, vcf_out: VCF = None, **kwargs) -> int:
        res = super().emit(vcf_out=vcf_out, **kwargs)

        if psnf_out := kwargs.get('psnf_out'):
            psnf_out.add_result(self)

        return res


class ErrorResult:
    error = True

    def __init__(self, msg: str):
        self.message = msg

    def __str__(self):
        return self.message

    def emit(self, **kwargs) -> int:
        log.error(f'{self.message}')
        return 0



#!/usr/bin/env python3
#
# Sniffles2
# A fast structural variant caller for long-read sequencing data
#
# Created:     14.01.2024
# Author:      Hermann Romanek
# Maintainer:  Hermann Romanek
# Contact:     sniffles@romanek.at
#
"""
Resource monitoring utilities (currently memory) for Sniffles
"""
import logging
import os
import time
from dataclasses import dataclass
from threading import Thread

from sniffles.config import SnifflesConfig


log = logging.getLogger('sniffles.ResourceMonitor')

@dataclass
class MemoryUsage:
    """
    A snapshot of memory usage for this process and all its workers
    """
    pid: int
    mem: float  # RSS in KiB


class ResourceMonitor:
    _pid: int
    _parent: float
    _unknown: float
    _workers: dict[int, MemoryUsage]
    _filename = None

    WORKER_MEMORY_LIMIT = 1024 * 1024 * 2  # in KiB

    def __init__(self, config: SnifflesConfig):

        self._slots = config.threads
        self._update_interval = config.dev_monitor_memory or 30.0
        self._pid = os.getpid()

        try:
            from psutil import Process
        except ImportError:
            logging.getLogger('sniffles.dependencies').warning(f'Dependency psutil not available - resource monitoring is disabled.')
            self._process = None
            self._workers = {}
        else:
            self._process = Process(self._pid)

            self._running = True

            if config.dev_monitor_memory:
                self._filename = filename = f'memory-{config.run_id}.csv'
                try:
                    self._file = open(filename, "w")
                    self._file.write(self._generate_header())
                except:
                    log.exception(f'Unable to write memory usage to {filename}')
                else:
                    log.info(f'Logging memory usage to {filename}')
            else:
                self._file = None

            self.update()
            self._update_thread = Thread(target=self._run_update, daemon=True)
            self._update_thread.start()

    @property
    def parent(self) -> float:
        return self._parent

    @property
    def workers(self) -> dict[int, float]:
        """
        dict pid -> memory
        """
        return {
            obj.pid: obj.mem for obj in self._workers.values() if obj is not None
        }

    @property
    def total(self) -> float:
        return self.parent + sum(self.workers.values())

    def stop(self):
        """
        Stop memory monitoring thread. This may take up to update_interval to complete.
        """
        self._running = False

    def _generate_header(self):
        return f'''total,pidp,memp,{','.join(f'pid{i},mem{i}' for i in range(self._slots))}\n'''

    def update(self):
        self._parent = self._process.memory_info().rss / 1024.0
        self._workers = {
            child.pid: MemoryUsage(child.pid, child.memory_info().rss / 1024.0) for child in self._process.children()
        }

        if self._file:
            self._file.write(f'''{','.join(self.generate())}\n''')
            self._file.flush()

        # print(f'Memory usage: {self.total / 1024:.2f} KiB')

    def _run_update(self):
        while self._running:
            self.update()
            time.sleep(self._update_interval)

        if self._file:
            self._file.close()

    def generate(self):
        yield str(self.total)
        yield str(self._pid)
        yield str(self.parent)
        for k, v in self.workers.items():
            yield str(k)
            yield str(v)

    @property
    def filename(self) -> str:
        return self._filename

    def __call__(self, worker_id: int, worker_pid: int) -> bool:
        """
        Call the resource monitor to get a recycle hint. Returns True if the given process should recycle.
        """
        if worker_pid in self._workers:
            return self._workers[worker_pid].mem > self.WORKER_MEMORY_LIMIT

        return False



from setuptools import setup, find_packages

setup(
    name='sniffles',
    version='2.6.0',
    packages=find_packages(),
    url='https://github.com/fritzsedlazeck/Sniffles',
    license='MIT',
    author='Moritz Smolka, Hermann Romanek',
    author_email='moritz.g.smolka@gmail.com, sniffles@romanek.at',
    description='A fast structural variation caller for long-read sequencing data',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
)



#!/usr/bin/env python3
#
# Sniffles2
# A fast structural variant caller for long-read sequencing data
#
# Created:     09.09.2021
# Author:      Moritz Smolka
# Maintainer:  Hermann Romanek
# Contact:     sniffles@romanek.at
#

from dataclasses import dataclass
import collections

from sniffles import util


@dataclass
class Group:
    """
    A sequence to which other sequences are pseudoaligned (mappings) through exact kmer matches (anchors)
    Can be reduced to a consensus sequence using the flatten() method
    """
    anchors: dict
    mappings: dict
    consensus_min: int
    consensus_max: int
    start_len: int
    origin: object
    support: list

    def flatten(self, frameshift_origin=False, return_debug=False):
        override_origin_min = 3
        override_origin_indel_ratio = 0.6

        consensus = ""
        origin_space_consensus = ""
        ins_consensus = ""
        del_consensus = ""
        mappings_consensus = ""
        skip_until = self.consensus_min - 1
        for ci in range(self.consensus_min, self.consensus_max + 1, 1):
            if ci not in self.mappings:
                origin_space_consensus += "-"
                ins_consensus += "-"
                del_consensus += "-"
                mappings_consensus += "-"
                continue
            type_fragments = {"M": [], "I": [], "D": []}
            type_fragments_counts = {"M": 0, "I": 0, "D": 0}
            for maptype, maplen, map_last_i, map_i, map_j, map_lead in self.mappings[ci]:
                type_fragments[maptype].append((maplen, map_lead.seq[map_last_i:map_i]))
                type_fragments_counts[maptype] += 1

            type_consensus = sorted(type_fragments, key=lambda t: len(type_fragments[t]))[-1]
            assert (len(type_fragments[type_consensus]) > 0)

            # Favor match over I/D when ratio is below 2

            if type_consensus == "I" and type_fragments_counts["M"] > 0 and type_fragments_counts["I"] / float(type_fragments_counts["M"] + type_fragments_counts["I"]) < override_origin_indel_ratio:
                type_consensus = "N"

            if type_consensus == "D" and type_fragments_counts["M"] > 0 and type_fragments_counts["D"] / float(type_fragments_counts["M"] + type_fragments_counts["D"]) < override_origin_indel_ratio:
                type_consensus = "N"

            if type_consensus == "M":
                # print(type_fragments_seq)
                if len(type_fragments[type_consensus]) < 3:
                    # Favor original sequence when only two mappings present
                    maplen, seq = type_fragments[type_consensus][0]
                    common_base, count = seq[0], 1
                else:
                    common_base, count = collections.Counter(seq[0] for maplen, seq in type_fragments[type_consensus]).most_common(1)[0]

                if ci > skip_until:
                    consensus += common_base
                    origin_space_consensus += common_base
                else:
                    origin_space_consensus += "-"
                ins_consensus += "-"
                del_consensus += "-"
                mappings_consensus += str(min(9, count))

            elif type_consensus == "N":
                # Favor original sequence when only two mappings present
                maplen, seq = type_fragments["M"][0]
                common_base, count = seq[0], 1

                if ci > skip_until:
                    consensus += common_base
                    origin_space_consensus += common_base
                else:
                    origin_space_consensus += "-"
                ins_consensus += "-"
                del_consensus += "-"
                mappings_consensus += str(min(9, count))

            elif frameshift_origin and type_consensus == "I":
                common_maplen, count = collections.Counter(maplen for maplen, seq in type_fragments[type_consensus]).most_common(1)[0]
                common_seqlen, count = collections.Counter(len(seq) for maplen, seq in type_fragments[type_consensus] if maplen == common_maplen).most_common(1)[0]
                origin_space_consensus += "^"
                ins_consensus += str(min(9, maplen))
                del_consensus += "-"
                mappings_consensus += str(min(9, count))

                selected = [seq for maplen, seq in type_fragments[type_consensus] if common_maplen == maplen and len(seq) == common_seqlen]

                for i in range(common_seqlen):
                    common_ins_base, _ = collections.Counter(seq[i] for seq in selected).most_common(1)[0]
                    consensus += common_ins_base

                skip_until = ci + common_seqlen + common_maplen - 1

            elif frameshift_origin and type_consensus == "D":
                del_consensus = del_consensus[:-1] + str(min(9, maplen))

        if not return_debug:
            return consensus
        else:
            return origin_space_consensus, ins_consensus, del_consensus, mappings_consensus, consensus


def make_group(lead, klen, skip, repetitive_kmers):
    anchors = {}
    for i, kmer in iter_kmers(lead.seq, klen, skip):
        if kmer in repetitive_kmers:
            continue
        if not kmer in anchors:
            anchors[kmer] = [i]
        else:
            anchors[kmer].append(i)

    return Group(anchors=anchors,
                 mappings={i: [("M", 1, i, i + 1, i, lead)] for i in range(len(lead.seq))},
                 consensus_min=0,
                 consensus_max=len(lead.seq) - 1,
                 start_len=len(lead.seq),
                 origin=lead,
                 support=[lead])


def iter_kmers(seq, klen, skip):
    for i in range(0, len(seq) - klen, skip):
        yield (i, seq[i:i + klen])


def from_leads(leads, initial_lead, max_grp=3, klen=15, skip=1, skip_repetitive=1, extend=False):
    # TODO: Handle DELs in flattening
    # TODO: Allow mapping across putative INDELs at sequence ends, allowing correcting them
    # TODO: Dynamically adapt kmer len threshold so short+long origs can be assembled at the same time, same for repetitive_limit
    # TODO: Ensure that kmers before and after last mapping are added as (secondary) anchors (hints 4 indels)

    repetitive_limit = len(leads) + 2

    # Generate repetitive kmer list
    kmer_counts = {}
    # for lead in leads:
    for i, kmer in iter_kmers(initial_lead.seq, klen=klen, skip=skip_repetitive):
        if kmer not in kmer_counts:
            kmer_counts[kmer] = 1
        else:
            kmer_counts[kmer] += 1
    repetitive_kmers = set(kmer for kmer, count in kmer_counts.items() if count > repetitive_limit)

    # Create initial group
    leads = sorted(leads, key=lambda ld: len(ld.seq))
    groups = [make_group(initial_lead, klen, skip, repetitive_kmers)]

    for lead in leads:
        if lead.seq is None:
            continue

        assigned = False

        # Try to place this lead's INS seq in the group
        for group in groups[:max_grp]:
            mapped = False
            last_i, last_j, last_was_repeat = None, None, False
            first_i, first_j = None, None

            # Attempt to anchor kmers from seq to current group
            for i, kmer in iter_kmers(lead.seq, klen=klen, skip=skip):
                if kmer not in group.anchors:  # and len(set(group.anchors[kmer]))==1:
                    continue

                if len(set(group.anchors[kmer])) > 1:
                    if last_j is None:
                        # Do not allow the first kmer to have multiple potential positions
                        continue

                    # Kmer occurs at multiple points in group sequence. Attempt to choose anchor most suitable for linear mapping
                    j = group.anchors[kmer][0]
                    for anchor_j_curr in group.anchors[kmer][1:]:
                        if abs(anchor_j_curr - (last_j + 1)) < abs(j - (last_j + 1)):
                            j = anchor_j_curr
                    last_was_repeat = True
                else:
                    # Kmer can be unambiguously placed on group sequence
                    j = group.anchors[kmer][0]

                if first_i is None and first_j is None:
                    first_i = i
                    first_j = j

                if last_i is not None and last_j is not None:
                    # Determine mapping type by position of current and last kmer on the lead sequence and group sequence
                    dist_lead = i - last_i
                    dist_group = j - last_j

                    if not last_was_repeat and dist_lead > dist_group:
                        maptype = "D"
                        maplen = dist_group - dist_lead
                        if not last_j in group.mappings:
                            group.mappings[last_j] = []
                        group.mappings[last_j].append((maptype, maplen, last_i, i, j, lead))

                    elif not last_was_repeat and dist_lead < dist_group:
                        maptype = "I"
                        maplen = dist_group - dist_lead
                        if not last_j in group.mappings:
                            group.mappings[last_j] = []
                        group.mappings[last_j].append((maptype, maplen, last_i, i, j, lead))

                    else:
                        maptype = "M"
                        maplen = last_i - i
                        act_j = last_j
                        for act_i in range(last_i, i):
                            if act_j not in group.mappings:
                                group.mappings[act_j] = []
                            group.mappings[act_j].append((maptype, maplen, act_i, act_i + 1, last_j, lead))
                            act_j += 1

                mapped = True
                last_i = i
                last_j = j
                last_was_repeat = False

            if mapped:
                group.support.append(lead)
                assigned = True
                max_i = len(lead.seq)

                # Extend consensus group leftwards
                if extend and first_i - (first_j - group.consensus_min) > 0:  # and first_j < 1.5*klen:
                    for i in range(first_i, -1, -1):
                        curr_j = first_j - (first_i - i)
                        if curr_j <= group.consensus_min:  # +klen/1:
                            assert (curr_j == group.consensus_min)
                            kmer = lead.seq[i:i + klen]
                            if kmer not in repetitive_kmers:
                                group.anchors[kmer] = [curr_j]
                            group.consensus_min -= 1
                            if curr_j not in group.mappings:
                                group.mappings[curr_j] = []
                            group.mappings[curr_j].append(("M", 1, i, i + 1, first_j, lead))

                # Extend consensus group rightwards
                if extend and last_j + (max_i - last_i) > group.consensus_max:
                    for i in range(last_i, max_i, 1):
                        curr_j = last_j + (i - last_i)
                        if curr_j + klen >= group.consensus_max:  # -klen/1:
                            assert (curr_j + klen == group.consensus_max)
                            kmer = lead.seq[i:i + klen]
                            if kmer not in repetitive_kmers:
                                group.anchors[kmer] = [curr_j]
                            group.consensus_max += 1
                            if curr_j not in group.mappings:
                                group.mappings[curr_j] = []
                            group.mappings[curr_j].append(("M", 1, i, i + 1, last_j, lead))
                break

        if not assigned:
            groups.append(make_group(lead, klen, skip, repetitive_kmers))
            # groups.sort(key=lambda g: len(g.mappings), reverse=True)

    return groups


def novel_from_reads(best_lead, other_leads, klen, skip, skip_repetitive, debug=False):
    consensus_min = 2
    maxshift = klen
    minspan = 0.2
    minalns = 0.25
    minident = 0.5
    minident_abs = 5
    minbestdiff = 3

    alignments = []
    anchors = {}
    taboo = set()
    for i, kmer in iter_kmers(best_lead.seq, klen=klen, skip=skip_repetitive):
        if kmer in taboo:
            continue
        if kmer in anchors:
            del anchors[kmer]
            taboo.add(kmer)
            continue
        anchors[kmer] = i

    for leadi, lead in enumerate(other_leads):
        last_i = None
        last_j = None
        conseq = ""
        span = 0
        for j, kmer in iter_kmers(lead.seq, klen=klen, skip=skip):
            if kmer not in anchors:
                continue
            i = anchors[kmer]
            if abs(i - j) > maxshift:
                continue
            if last_i is not None and i <= last_i:
                continue

            if last_i is None:
                if j > 0:
                    conseq = "-" * i
            else:
                fwd_i = i - last_i
                fwd_j = j - last_j
                if len(conseq) + fwd_j > len(best_lead.seq):
                    fwd_j = len(best_lead.seq) - len(conseq)

                if fwd_i == fwd_j and fwd_j > 0:
                    span += (j - last_j)
                    m = 0
                    for l in range(1, (j - last_j) + 1):
                        if lead.seq[last_j + l] == best_lead.seq[last_i + l]:
                            m += 1
                    ident = m / float((j - last_j))
                    if ident >= minident:
                        conseq += lead.seq[last_j:j][:fwd_j]
                    else:
                        conseq += "-" * (fwd_j)
                else:
                    conseq += "-" * (fwd_j)
            last_i = i
            last_j = j

        if len(conseq) < len(best_lead.seq):
            conseq += "-" * (len(best_lead.seq) - len(conseq))

        conseq_new = []
        h = 0
        while h < len(best_lead.seq):
            if conseq[h] == "-":
                conseq_new.append("-")
                h += 1
            else:
                buffer = []
                ident = 0
                while h < len(best_lead.seq) and conseq[h] != "-":
                    ident += (best_lead.seq[h] == conseq[h])
                    buffer.append(conseq[h])
                    h += 1
                if ident / float(len(buffer)) > minident and ident > minident_abs:
                    conseq_new.append("".join(buffer))
                else:
                    conseq_new.append("-" * len(buffer))
        conseq = "".join(conseq_new)

        if span / float(len(best_lead.seq)) > minspan:
            alignments.append(conseq)

    maxal = 1
    for i in range(len(best_lead.seq)):
        maxal = max(maxal, len([best_lead.seq[i]] + [a[i] for a in alignments if not a[i] in "^_"]))
    maxal = float(maxal)

    flattened = ""
    for i in range(len(best_lead.seq)):
        al = [a[i] for a in alignments if not a[i] == "-"]
        if len(al) < consensus_min or len(al) / maxal < minalns:
            flattened += best_lead.seq[i]
        else:
            top = util.most_common([best_lead.seq[i]] + al)
            if len(top) > 1 and top[0][0] - top[1][0] >= minbestdiff:
                flattened += top[0][1]
            else:
                flattened += best_lead.seq[i]

            # if top[0][0]/float(len(al)+1) < 0.75:
            #    flattened+=best_lead.seq[i]
            # else:
            #    flattened+=top[0][1]

    # print("FLT",flattened)
    # if debug:
    #    print("B",best_lead.seq)
    #    for o in other_leads:
    #        print("O",o.seq)
    #    print("F",flattened)
    #    print("=====")
    return flattened



import logging
import unittest
from typing import Tuple
from unittest.mock import patch, mock_open

from sniffles.config import SnifflesConfig
from sniffles.region import Region


class TestRegions(unittest.TestCase):
    """
    Tests --regions parameter
    """
    @staticmethod
    def _get_common_args() -> Tuple:
        """
        Get common required arguments for sniffles
        """
        return '--input', 'input.bam', '--vcf', 'out.vcf'

    def test_GoodFile(self):
        data = """
# comment line is ok
chr1\t100\t200\n
chr1\t500\t600\n
chr3\t500\t600\n
        """
        with patch("builtins.open", mock_open(read_data=data)) as mock_file:
            config = SnifflesConfig(*self._get_common_args(), '--regions', 'regions.bed')

        self.assertDictEqual(
            config.regions_by_contig,
            {'chr1': [Region('chr1', 100, 200), Region('chr1', 500, 600)], 'chr3': [Region('chr3', 500, 600)]}
        )
        mock_file.assert_called_with("regions.bed", "r")

    def test_ContigConflict(self):
        """
        Expect exception if both --contig and --regions are provided
        """
        with self.assertRaises(SystemExit):
            SnifflesConfig(*self._get_common_args(), '--regions', 'regions.bed', '-c', 'chr6')

    def test_FileNotFound(self):
        with self.assertRaises(FileNotFoundError):
            SnifflesConfig(*self._get_common_args(), '--regions', 'regions.bed')

    def test_FileInvalidLines(self):
        data = """
... <- invalid line
chr1\t100\t200\n  valid line

"""
        with self.assertLogs(level=logging.WARNING), patch("builtins.open", mock_open(read_data=data)) as mock_file:
            config = SnifflesConfig(*self._get_common_args(), '--regions', 'regions.bed')

        self.assertDictEqual(
            config.regions_by_contig,
            {'chr1': [Region('chr1', 100, 200)]}
        )
